%
% This is the third encoder script for
%
%   storebror by Pex 'Mahoney' Tufvesson, August 2011
%
% Run it using Matlab, you'll probably need a couple of toolboxes
% and a lot of patience.
% After running all matlab scripts, you'll find a number of
% files named "SID*.s".
% Copy these into the demo folder.
%
% Want to know more?  http://mahoney.c64.org
% Want to hear C64 a cappella?  Visa Rster at http://www.livet.se/visa
% Want a happy life? Do something _not_ requiring electricity! :-D
%

format long;
format compact;
clear;

load 'matlab_whichWaveformsAreChosen.mat' %The order of which waveforms were chosen
load 'matlab_waveformStatus.mat' %Which waveform every waveforms resembles most.
load 'matlab_SIDVoiceVolume.mat' %Play/noPlay
load 'matlab_Formant0.mat' %contains pitch in Formant0
load 'matlab_waveformsNormalizedFD.mat'
load 'matlab_waveforms.mat'
load 'matlab_filteredAudioX2.mat' % contains filteredAudioX
waveformsNormalized = waveformsNormalizedFD;
baseFreq = Formant0;
%[originalAudio,Fs,bits]=wavread('Suzanne_Vega_-_Toms_Diner_mono');

% Noteno, 1=55 Hz=A in octave 2
noteNo=(0:60);
twelfthRootOfTwo=2 .^ (1/12); % =1.059463094;
Hz = 55 * ( twelfthRootOfTwo .^ noteNo );

% Now, we have everything to "simulate" playback.
% SIDVoiceVolume tells us when to sound, and when to be quiet.
% Every time SIDVoiceVolume == 1, increase waveformNo
% Use waveformStatus to check which waveform we shall play.
% Increase pointer into SIDVoiceVolume every 8.192ms.


Fs=44100;
% We want to detect pitch in 100Hz.
sliceSecondsTarget = 1/100;
% This is due to using a 64-word buffer for
% playing samples, using a 7812.5Hz sample rate.
% So the audio can be sliced up in 8.192ms slices.
% Input audio is 44100Hz, 16-bit.
% Each slice is:
sliceLength = round(Fs * sliceSecondsTarget) %#ok<NOPTS>
% This is what we really got:
sliceSeconds = sliceLength / Fs %#ok<NOPTS>
% Calculate "how wrong" the speed will be due to rounding
% the sliceLength to an integer number of 16-bit audio samples.
timeSkew = (sliceSecondsTarget - sliceSeconds) / sliceSecondsTarget %#ok<NOPTS>

nofSlices=length(SIDVoiceVolume)


% Build a table with waveforms and SIDVoiceVolume combined, so that sliceNo
% can index it directly:
waveformIndex = 1;
sliceWavetableIndex = zeros(1,nofSlices);
for sliceNo=1:nofSlices-3,
   if (SIDVoiceVolume(sliceNo) == 1)
       waveformNoInWaveformsNormalized = waveformStatus(waveformIndex);
       % Find this waveform in whichWaveformsAreChosen, and renumber it:
       tableIndex = find(whichWaveformsAreChosen == waveformNoInWaveformsNormalized);
       sliceWavetableIndex(sliceNo) = tableIndex(1) - 1;
       waveformIndex = waveformIndex + 1;
   else
       sliceWavetableIndex(sliceNo) = 127;  %Silent waveform
%       waveformIndex = waveformIndex + 1;
   end
end

% Now, build the include file to use with all the waveforms in it:
SIDwaveformLength = 64
waveformSize = size(waveformsNormalized);
nofWaveforms = waveformSize(1);
waveformSize = waveformSize(2);

waveformsToChoose=128;

amplitude = 31;
bias = 32;
dist = 1.07;

SIDwaveforms = zeros(256,SIDwaveformLength);
for waveformNo = 0:(waveformsToChoose-1),
    waveformsNormalizedIndex = whichWaveformsAreChosen(waveformNo+1);
    if waveformsNormalizedIndex ~= 0
        for i = 0:SIDwaveformLength-1
            % Time to resample the waveforms so they fit into
            % SIDwaveformLength number of bytes.
            value = waveformsNormalized(waveformsNormalizedIndex, 1 + floor((i / SIDwaveformLength * waveformSize)));
            % Now, add some noise to it for reducing quantization noise:
%            value = value + rand(1) * 0.5 * (1 / 256);
            % And quantize it into 8 signed bits:
            value = round(value * amplitude * dist);
            value = max(-(amplitude),value); % limit it
            value = min(amplitude,value);  % limit it
            SIDwaveforms(waveformNo + 1,i + 1) = value;
        end
    end
end

% Dump the encoded SIDwaveforms into a file
% This file shall be placed at $xx00 in C64 memory
% This is the old way of dumping the waveforms in memory:
% fid = fopen('SIDwaveforms.s','w');
% fprintf(fid,'; encoded SIDwaveforms generated by cubase64_4encodeWaveforms.m\n');
% fprintf(fid,'; this table contains 256*%d = %d values\n',SIDwaveformLength,SIDwaveformLength*256);
% fprintf(fid,'; To be placed at even 256-byte boundary in C64 memory');
% fprintf(fid,'; Generated on the %s\n',date);
% fprintf(fid,'; Pex Mahoney Tufvesson, 2010\n');
% fprintf(fid,'\n');
% for i = SIDwaveformLength-1:-1:0
%     fprintf(fid,';These are byte number #%d for all the 256 waveforms:\n',i);
%     for waveformNo = 0:255,
%         fprintf(fid,'  .byte %d\n',SIDwaveforms(waveformNo+1,i+1) + 128);
%     end
% end
% fprintf(fid,'; End of generated data\n');
% fprintf(fid,'\n');
% fclose(fid);



fid = fopen('SIDwaveforms.s','w');
fprintf(fid,'; encoded SIDwaveforms generated by storebror_03.m\n');
fprintf(fid,'; this table contains 128*%d = %d values\n',SIDwaveformLength,SIDwaveformLength*256);
fprintf(fid,'; To be placed at an even 256-byte boundary in C64 memory');
fprintf(fid,'; Generated on the %s\n',date);
fprintf(fid,'; Pex Mahoney Tufvesson, 2011\n');
fprintf(fid,'\n');
for waveformNo = 0:(waveformsToChoose-1),
    fprintf(fid,';This is waveform number %d, forwards:\n',waveformNo);
    fprintf(fid,'wave%dstart:\n',waveformNo);
    for i = 0:SIDwaveformLength-1
        fprintf(fid,'  .byte %d\n',SIDwaveforms(waveformNo+1,i+1) + bias);
    end
    fprintf(fid,'wave%dend:\n',waveformNo);
end
fprintf(fid,'; End of generated data\n');
fprintf(fid,'\n');
fclose(fid);


fid = fopen('SIDwaveformPoiTable.s','w');
fprintf(fid,'; encoded SIDwaveformPointers generated by storebror_03.m\n');
fprintf(fid,'; To be placed at $xx00 in C64 memory');
fprintf(fid,'; Generated on the %s\n',date);
fprintf(fid,'; Pex Mahoney Tufvesson, 2011\n');
fprintf(fid,'\n');
fprintf(fid,'waveLength = %d\n',SIDwaveformLength);
fprintf(fid,'SIDwaveformPoiTableMSB:\n');
for waveformNo = 0:(waveformsToChoose-1),
    fprintf(fid,'  .byte >wave%dstart\n',waveformNo);
end
fprintf(fid,'\nSIDwaveformPoiTableLSB:\n');
for waveformNo = 0:(waveformsToChoose-1),
    fprintf(fid,'  .byte <wave%dstart\n',waveformNo);
end
fprintf(fid,'; End of generated table\n');
fprintf(fid,'\n');
fclose(fid);









% We shall dump all the separate lists with which wave we shall play.
% One byte per 10ms wave.

% First, divide the long vector into our separate files again.

nof_audio_files = 14;
si = size(sliceWavetableIndex);
each_file_sliceLength = si(2) / nof_audio_files;
tnof = 0;

for fileNo = 1:nof_audio_files,

    % First, try to find the "end index" of this file.
    % They all end with loads of silence, just discard that.
    
    startIndex = 1 + (fileNo-1)*each_file_sliceLength - 1;
    if startIndex < 1,
        startIndex = 1;
    end
    endIndex = (fileNo-1+1)*each_file_sliceLength - 2

    i = endIndex - 1;
    while sliceWavetableIndex(i) == 127,
        i = i - 1;
    end

    endIndex = i

    % Dump the wavelists into files
    filename = sprintf('SIDwavelist%d.s',fileNo);
    fid = fopen(filename,'w');
    fprintf(fid,'; encoded SIDwaveTable generated by storebror_03.m\n');
    fprintf(fid,'; this table contains %d values unpacked\n',endIndex-startIndex);
    fprintf(fid,'; Generated on the %s\n',date);
    fprintf(fid,'; Pex Mahoney Tufvesson, 2011\n');
    fprintf(fid,'\n');
    i = startIndex;
    nof = 0;
    while (i <= endIndex)
        % New: Insert wavelist packing here.
        % 129 = play next value two times
        % 130 = play next value three times (saves 1 byte mem)
        % 131 = play next value four times (saves 2 bytes of memory)
        thisi = sliceWavetableIndex(i);
        repeat_count = 0;
        while ((i+repeat_count+1 <= endIndex) && (sliceWavetableIndex(i+repeat_count+1) == thisi))
            repeat_count = repeat_count + 1;
        end
        if (repeat_count <= 2)
            fprintf(fid,'  .byte %d\n',sliceWavetableIndex(i));
            i = i + 1;
            nof=nof+1;
        else
            fprintf(fid,'  .byte %d\n',repeat_count+128);
            fprintf(fid,'  .byte %d\n',sliceWavetableIndex(i));
            i = i + repeat_count + 1;
            nof=nof+2;
        end
    end
    if (fileNo ~= nof_audio_files),
        % Insert the echo (silence at the end of the phrase) but only
        % for single channel vers-parts, not for the 3-channel ref.
        fprintf(fid,'  .res 1,127\n');       % This is the echo length
    end
    fprintf(fid,'; this table contained %d values packed\n',nof);
    fprintf(fid,'; End of generated data\n');
    fprintf(fid,'\n');
    fclose(fid);

    tnof = tnof + nof;
end


die


%vers1a: ;14*24 = 336 renders
%  .res 14*24,0
%  .byte $ff
%vers1b: ;18*24 = 432 renders
%  .res 18*24,0
%  .byte $ff





% Dump the encoded SIDencWave into a file
fid = fopen('SIDencWave.s','w');
fprintf(fid,'; encoded SIDwaveTable generated by cubase64_4encodeWaveforms.m\n');
fprintf(fid,'; this table contains %d values\n',SIDencWaveIndex);
fprintf(fid,'; Generated on the %s\n',date);
fprintf(fid,'; Pex Mahoney Tufvesson, 2010\n');
fprintf(fid,'\n');
for i = 1:SIDencWaveIndex
    fprintf(fid,'  .byte %d\n',SIDencWave(i));
end
fprintf(fid,'; End of generated data\n');
fprintf(fid,'\n');
fclose(fid);







for index=1:SIDindex,
   if (SIDwave(index) ~= 255)
       SIDencSpeed(SIDencSpeedIndex) = SIDspeed(index);
       SIDencSpeedIndex = SIDencSpeedIndex + 1;
   end
end
%SIDencSpeed = SIDencSpeed(1:SIDencSpeedIndex);
SIDencSpeedIndex
SIDencSpeedIndex = SIDencSpeedIndex-2

figure(51);
plot(SIDencWave);
title('SIDencWave');
figure(52);
plot(SIDencSpeed);
title('SIDencSpeed');

% SIDencSpeed could be reasonably cheap encoded by encoding
% "differences" from -7 to +7 with one nybble.
% If nybble = -8, then grab next byte (may be split in 
% two separate nybbles), and put that into Speed value.
% First byte is the start speed.

SIDenc2Speed = zeros(1,SIDencSpeedIndex*2);
SIDenc2Speed(1) = floor(SIDencSpeed(1) / 16); % Split nybbles
SIDenc2Speed(2) = SIDencSpeed(1) - floor(SIDencSpeed(1) / 16) * 16;     % Split nybbles
oldSpeed = SIDencSpeed(1);
SIDenc2SpeedIndex = 3;

for index=1:SIDencSpeedIndex,
    if (abs(SIDencSpeed(index) - oldSpeed) > 7)
        SIDenc2Speed(SIDenc2SpeedIndex) = 8; % Use next 8 bits for speed
        SIDenc2SpeedIndex = SIDenc2SpeedIndex + 1;
        SIDenc2Speed(SIDenc2SpeedIndex) = floor(SIDencSpeed(index) / 16);
        SIDenc2SpeedIndex = SIDenc2SpeedIndex + 1;
        SIDenc2Speed(SIDenc2SpeedIndex) = SIDencSpeed(index) - floor(SIDencSpeed(index) / 16) * 16;
        SIDenc2SpeedIndex = SIDenc2SpeedIndex + 1;
        oldSpeed = SIDencSpeed(index);
    else
        SIDenc2Speed(SIDenc2SpeedIndex) = SIDencSpeed(index) - oldSpeed;
        if (SIDenc2Speed(SIDenc2SpeedIndex) < 0)
            SIDenc2Speed(SIDenc2SpeedIndex) = SIDenc2Speed(SIDenc2SpeedIndex) + 16;
        end
        SIDenc2SpeedIndex = SIDenc2SpeedIndex + 1;
        oldSpeed = SIDencSpeed(index);
    end
end

SIDenc2Speed = SIDenc2Speed(1:SIDenc2SpeedIndex+3);
SIDenc2Length = SIDenc2SpeedIndex/2

% Now, combine all gathered nybbles:
SIDenc3Speed = zeros(1,floor(SIDenc2SpeedIndex/2) + 2);
for index=0:(floor(SIDenc2SpeedIndex/2)),
    SIDenc3Speed(index+1) = SIDenc2Speed(index*2 + 1) * 16 + SIDenc2Speed(index*2+2);
end

SIDenc3SpeedLength = floor(SIDenc2SpeedIndex/2) + 1

figure(53)
plot(SIDenc3Speed);
title('SIDenc3Speed');

% Dump the encoded SIDenc3Speed into a file
fid = fopen('SIDenc3Speed.s','w');
fprintf(fid,'; encoded SIDspeed generated by cubase64_4encodeWaveforms.m\n');
fprintf(fid,'; this table contains %d values\n',SIDenc3SpeedLength);
fprintf(fid,'; Generated on the %s\n',date);
fprintf(fid,'; Pex Mahoney Tufvesson, 2010\n');
fprintf(fid,'\n');
for i = 1:SIDenc3SpeedLength
    fprintf(fid,'  .byte %d\n',SIDenc3Speed(i));
end
fprintf(fid,'; End of generated data\n');
fprintf(fid,'\n');
fclose(fid);



%The note values used through the whole song are:
% Note Name   Hz                     nofSamples@7812.5Hz
% D   Hz(18)  146.8323839587038 Hz   53.206927445904853
% E   Hz(20)  164.8137784564350 Hz   47.401983457742659
% F   Hz(21)  174.6141157165020 Hz   44.741514555925860
% F#  Hz(22) 184.9972113558173 Hz    42.230366299812502
% G#  Hz(24) 207.6523487899727 Hz    37.622979203099938
% A   Hz(25) 220 Hz                  35.511363636363598
% B   Hz(27) 246.9416506280622 Hz    31.637028343051789
% C#  Hz(29) 277.1826309768723 Hz    28.185387996594418
% D   Hz(30) 293.6647679174077 Hz    26.603463722952405





% When all waveforms are chosen, it's time to try to map _every_ single
% waveform that we have onto a linear combination of two waveforms.
% Maybe this will be with some heavy limitations due to memory
% requirements.
% And also, the number of volume levels will be only a few, probably
% only 25%, 33%, 50%, 66%, 75% and 100%.



% Let's try to create one waveform and modify this.

% The replay routine will have to take care of the pitching.
% And no pitch tables, we need to have "human" qualities of the
% pitch curve, so it needs to be imperfect.

% The mixing routine will blend "the old waveform" with a new one.
% The mixing routine will use pha to store values.
% The mixing routine cannot use pla to get values, since any
% interrupt that occurs will destroy the waveform.

% Or, the waveform could be stored, double buffered, in ZeroPage.


% Replay with no mixing routine:

% IRQ handling                    7
%         STA     SaveAnmi+1      3
%         ...a few clocks sync 0 - 4 of them
%         JMP     ZeroCode        3
%ZeroCode:
%        LDA     #$11             2
%        STA     $D404            4
%        LDA     #9               2
%        STA     $D404            4
%Poi:
%        LDA     $3000            4
%        STA     $D401            4
%        LDA     #1               2
%        STA     $D404            4
%        INC     Poi+1            5
%SaveAnmi:
%        LDA     #$A              2
%        JMP     $DD0C            3
%DD0C:   RTI                      6
%                               =55-59 clocks in all
% We have 126-55 = 67-71 clocks left per sample for the mixing routine
% A badline will take 40 of these. = 27-31 clocks left on a badline

